////////////////////////////////////////////////////////////////////////////////////////
// Module name: Client.cpp
// Project: PowerMonitor #42
// Module: MQTT client, based on PubSubClient library for Arduino.
// by Jan Breemer 
// by Jan Breemer  jan@breem.nl, janbee@hack42.nl// 
//
// Functions in this module:
//
// InitTopics ()
// TopicReceived ()
// UpdateClient ()
// InitClient ()
////////////////////////////////////////////////////////////////////////////////////////

#include "Globals.h"
#include <SPI.h>
#include <Ethernet.h>
#include <PubSubClient.h>
#include <string.h>

EthernetClient ethClient;
byte mac[] = {0xDE, 0xED, 0xBA, 0xFE, 0xFE, 0xED };
IPAddress ClientIP (192, 168, 2, 5);
IPAddress BrokerIP (192, 168, 2, 4);
PubSubClient Client(ethClient);

byte HeartBeats, Relays, RequestLeds;
bool DoPublishStatus, DoPublishHeartBeatAck;
bool ConnectionError, SubscriptionError;

// to publish:
char Topic_Status[40];
char Topic_Power[40];
char Topic_AirQuality[80];
char Topic_DallasTemperatures[80];
char Topic_HeartBeatAck[50];

// to subscribe to:
char Topic_SetRelays[40];
char Topic_SetRequestLights[40];
char Topic_HeartBeat[30];
char Payload[110];

// ***********************************************************************************
void InitTopics (void)
{
  // Topics to Publish:
  sprintf (Topic_Status, "%s", CLIENTID);
  strcat  (Topic_Status, "/Status");
  
  sprintf (Topic_Power, "%s", CLIENTID);
  strcat  (Topic_Power, "/Power");
  
  sprintf (Topic_AirQuality, CLIENTID);
  strcat  (Topic_AirQuality, "/AirQuality");

  sprintf (Topic_DallasTemperatures, CLIENTID);
  strcat  (Topic_DallasTemperatures, "/DallasTemperatures");
  
  sprintf (Topic_HeartBeatAck, CLIENTID);
  strcat  (Topic_HeartBeatAck, "/HeartBeatAck");
  
  // Topics to Subscribe to:
  #define Mediator "Mediator/"

  sprintf (Topic_SetRelays, CLIENTID);
  strcat  (Topic_SetRelays, "/SetRelays");

  sprintf (Topic_SetRequestLights, CLIENTID);
  strcat  (Topic_SetRequestLights, "/SetRequestLights");
  
  sprintf (Topic_HeartBeat, Mediator);
  strcat  (Topic_HeartBeat, "HeartBeat");

  //sprintf (Topic_HeartBeat, "#");
  //strcat  (Topic_HeartBeat, "HeartBeat");


  Serial.println ("Prepare Topics to publish: ");

  Serial.println (Topic_Status); 
  
  Serial.print (Topic_Power); 
#ifndef USE_CURRENTS 
  Serial.println (" not used"); 
#else 
  Serial.println ();  
#endif
     
  Serial.print (Topic_AirQuality); 
#if !defined USE_BME280 && !defined USE_CCS811 && !defined USE_SDS011 
  Serial.println (" not used");     
#else 
  Serial.println ();  
#endif
     
  Serial.print(Topic_DallasTemperatures);     
#ifndef USE_DALLAS_SENSORS 
  Serial.println (" not used"); 
#else
  Serial.println ();
#endif  
  Serial.println (Topic_HeartBeatAck); 
  Serial.println ();
    
  Serial.println ("Prepare Topics to subscribe to: "); 
  Serial.println (Topic_SetRelays);
  Serial.println (Topic_SetRequestLights);
  Serial.println (Topic_HeartBeat); 
  Serial.println ();
}

// ***********************************************************************************
void TopicReceived (char* Topic, char* Payload, unsigned int Length) 
{
  if (DiagnoseMode) Serial.print ("Message arrived: ");
  //Serial.print (" Topic= "); 
  //Serial.println (Topic);
  //Serial.print (" PayLoad= "); 
  //Serial.println (Payload);
   
  if (strcmp (Topic, Topic_HeartBeat) == 0)
  {
    sscanf (Payload, "%3d", &HeartBeats);
    DoPublishHeartBeatAck = true;
    if (DiagnoseMode)
    {      
      Serial.print ("HeartBeat ");
      Serial.print (" PayLoad= ");
      Serial.println (HeartBeats);
    }  
    HEARTBEATLED_TOGGLE
  }
  
  if (strcmp(Topic, Topic_SetRelays) == 0)
  {
    sscanf (Payload, "%3d", &Relays);
    if (DiagnoseMode)
    {      
      Serial.print ("SetRelays ");
      Serial.print (" PayLoad= ");
      Serial.println (Relays);
    }  
    PORTC = (Relays & ~0x01) | (PINC & 0x01); // don't touch bit 0 
  }
  
  if (strcmp(Topic, Topic_SetRequestLights) == 0)
  {
    sscanf (Payload, "%3d", &RequestLeds);
    if (DiagnoseMode)
    {      
      Serial.print ("SetReQuestLights ");
      Serial.print (" PayLoad= ");
      Serial.println (RequestLeds);
    }  
  }  
}

// ***********************************************************************************
void UpdateClient (void)
{
  Client.loop();
  if (ConnectionError || SubscriptionError) return;
  
  if (DoPublishHeartBeatAck) 
  { 
    DoPublishHeartBeatAck = false;
    sprintf (Payload, "%3d", HeartBeats);
    Client.publish (Topic_HeartBeatAck, Payload);  
    PUBLED_TOGGLE
    if (DiagnoseMode)
    { 
      Serial.print ("Publish T= ");
      Serial.print (Topic_HeartBeatAck);
      Serial.print (" P= ");
      Serial.println (Payload);
    }  
  }

  if (DoPublishStatus) 
  { 
    DoPublishStatus = false;
    Status = 0;
    if (InitialisationError) Status |= 0x80;
    if (RequestButton1Pressed) Status |= 0x01;
    if (RequestButton2Pressed) Status |= 0x02;
    sprintf (Payload, "%3d %s", Status, VersionString);
    Client.publish (Topic_Status, Payload);  
    PUBLED_TOGGLE
    if (DiagnoseMode)
    { 
      Serial.print ("Publish T= ");
      Serial.print (Topic_Status);
      Serial.print (" P= ");
      Serial.println (Payload);
    }  
  }

#ifdef USE_CURRENTS
  if (DoPublishPower) 
  { 
    DoPublishPower = false;
    Client.publish (Topic_Power, Payload); 
    PUBLED_TOGGLE
    if (DiagnoseMode)
    { 
      Serial.print ("Publish T= ");
      Serial.print (Topic_Power);
      Serial.print (" P= ");
      Serial.println (Payload);
    }
  }
#endif

  #if defined USE_BME280 || defined USE_CCS811 || defined USE_SDS011  
  if (DoPublishAirQuality)
  {
    DoPublishAirQuality = false;
  #ifndef USE_BME280
    Temperature = Hygro = Baro= -999;
  #endif
  #ifndef USE_CCS811
    eCO2 = eVOC = -999;
  #endif
  #ifndef USE_SDS011
    P10 = P2p5 = -999;
  #endif   
    sprintf (Payload, "%5d%5d%5d%5d%5d%5d%5d", Temperature, Hygro, Baro, eCO2, eVOC, P10, P2p5);
    Client.publish (Topic_AirQuality, Payload);
    PUBLED_TOGGLE
    if (DiagnoseMode)
    { 
      Serial.print ("Publish T= ");
      Serial.print (Topic_AirQuality);
      Serial.print (" P= ");
      Serial.println (Payload);
    }
  }
#endif
  
#ifdef USE_DALLAS_SENSORS
  if (DoPublishDallasTemperatures)
  {
    DoPublishDallasTemperatures = false;
    sprintf (Payload, "%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d", DallasTemperatures[0], 
      DallasTemperatures[1], DallasTemperatures[2], DallasTemperatures[3],
      DallasTemperatures[4], DallasTemperatures[5], DallasTemperatures[6], 
      DallasTemperatures[7], DallasTemperatures[8], DallasTemperatures[9]);
    Client.publish (Topic_DallasTemperatures, Payload);
    PUBLED_TOGGLE
    if (DiagnoseMode)
    { 
      Serial.print ("Publish T= ");
      Serial.print (Topic_DallasTemperatures);
      Serial.print (" P= ");
      Serial.println (Payload);
    }
  }
#endif  
}

// *********************************************************************************** 
void InitClient (void)
{
  bool SubscriptionsOK = true;
  Ethernet.begin (mac, ClientIP);
  Client.setServer (BrokerIP, 1883);
  Client.setCallback (&TopicReceived);
  Serial.print ("Client IP = ");
  Serial.println (ClientIP);
  Serial.print ("Broker IP = ");
  Serial.println (BrokerIP);
  Serial.println();

  Sequence_MQTT_Leds ();

  #ifdef USE_REQUEST_BOX1
    Sequence_Request1_Leds ();
  #endif  
  #ifdef USE_REQUEST_BOX2
    Sequence_Request2_Leds ();
    Serial.println();
  #endif 

  InitTopics();
  Serial.println("Try to connect to the MQTT Broker");
  if (Client.connect(CLIENTID)) 
  {
    bool result;
    // connection succeeded
    Serial.print(CLIENTID);
    Serial.println(" is now connected to the MQTT Broker");
    CONNLED_ON

    Serial.print("Subscribe to Topic: "); 
    Serial.print(Topic_SetRelays);  
    result = Client.subscribe(Topic_SetRelays);
    if (result) Serial.println("....OK");
    else Serial.println("....FAILED");
    if (!result) SubscriptionsOK = false;
    
    Serial.print("Subscribe to Topic: "); 
    Serial.print(Topic_SetRequestLights);  
    result = Client.subscribe(Topic_SetRequestLights);
    if (result) Serial.println("....OK");
    else Serial.println("....FAILED");
    if (!result) SubscriptionsOK = false;
    
    Serial.print("Subscribe to Topic: "); 
    Serial.print(Topic_HeartBeat);  
    result = Client.subscribe(Topic_HeartBeat);
    if (result) Serial.println("....OK");
    else Serial.println("....FAILED");
    if (!result) SubscriptionsOK = false;  
    if (SubscriptionsOK) SUBLED_ON 
    else 
    {
      SubscriptionError = true;
      Serial.println("There were errors while subscribing to Topics");
    }
  } 
  else 
  {
    ConnectionError = true;
    Serial.print("Connection WITH MQTT Broker failed. Client.state= ");
    Serial.println(Client.state());
    // mqttClient.state() will provide more information on why it failed.
  }
  Serial.println();
}
